#include <iostream>

// This simple example solves the following constrainted optimization problem:
//
// argmin_{x} x^T S x 
// s. t.  n/2 - x^T P x <= 0
//        n/2 - x^T Q x <= 0
//        -1 <= x_j <= 1, j=1,...,n 
//
// where n is the number of design variables, S, P, 
// and Q are symmetric n x n matrices with elements given by
//  s_{ij} = 2 + sin(4*pi*a_{ij}) / (1 + |i-j|)*ln(n),
//  p_{ij} = 1 + 2a_{ij} / (1+|i-j|)ln(n),
//  q_{ij} = 3 - 2a_{ij} / (1+|i-j|)ln(n),
// with a_{ij} = i+j-2 / (2n-2) for all i and j.

// The API of the optimizer is entirely defined in the conveks header
#include <conveks.h>

int main(int argc, char **argv)
{
  // Maximum number of iterations
  int maxIter = 20;

  // number of variables and constraints
  int numVariables = 100;

  std::vector<double> Smat(numVariables*numVariables);
  std::vector<double> Pmat(numVariables*numVariables);
  std::vector<double> Qmat(numVariables*numVariables);
  double lnn = (double)(1.0/std::log(numVariables));
  const double pi = std::acos(-1);
  for(unsigned int i = 1; i < numVariables+1; i++){
    for(unsigned int j = 1; j < numVariables+1; j++){
      double dincr = (double)((i+j-2.0)/(numVariables-1));
      //std::cout << "abs(ij): " << std::abs((int)(i-j)) << std::endl;
      Smat[(i-1)*numVariables+j-1] = lnn*(2.0+std::sin(2.0*pi*dincr))/(1+std::abs((int)(i-j)));
      Pmat[(i-1)*numVariables+j-1] = lnn*(1.0+dincr)/(1.0+std::abs((int)(i-j)));
      Qmat[(i-1)*numVariables+j-1] = lnn*(3.0-dincr)/(1.0+std::abs((int)(i-j)));
    }
  }

  // initial value of the variables in the design space
  std::vector<double> initialPoint(numVariables, 0.5);

  // lower bound for design variables (all set to 1)
  std::vector<double> lowerBound(numVariables, -1.0);

  // upper bound for design variables (all set to 10)
  std::vector<double> upperBound(numVariables, 1.0);

  // Initilize conveks
  conveks::initialize();

  // Initialize the MMA optimizer
  conveks::mma::initialize(initialPoint, lowerBound, upperBound);

  // Set some options for MMA
  conveks::mma::option::setNumber("General.Verbosity", 4);
  conveks::mma::option::setNumber("SubProblem.asymptotesRmax", 10.0);
  conveks::mma::option::setNumber("SubProblem.asymptotesRmin", 0.01);

  // Get iteration count (here it will be 1 - could be different in case of restart)
  int it;
  conveks::mma::getOuterIteration(it);

  while(it <= maxIter){

    // get (copy of) current point
    std::vector<double> x;
    conveks::mma::getCurrentPoint(x);

    std::vector<double> sx(numVariables);
    std::vector<double> px(numVariables);
    std::vector<double> qx(numVariables);
    for(unsigned int i = 0; i < numVariables; i++){
      sx[i] = 0.0;px[i] = 0.0;qx[i] = 0.0;
      for(unsigned int j = 0; j < numVariables; j++){ 
        sx[i] += Smat[i*numVariables+j] * x[j];
        px[i] += Pmat[i*numVariables+j] * x[j];
        qx[i] += Qmat[i*numVariables+j] * x[j];
      }
    }

    // compute objective function and constraints
    // at point `x' in the design space
    double objective=0.0;
    std::vector<double> constraints(2);
    for(unsigned int i = 0; i < numVariables; i++){
      objective += sx[i] * x[i];
      constraints[0] -= px[i]*x[i]; 
      constraints[1] -= qx[i]*x[i]; 
    }
    constraints[0] += numVariables*0.5; 
    constraints[1] += numVariables*0.5; 
  
    // as well as their sensitivity with respect to design variables at `x'
    std::vector<double> grad_objective(numVariables);
    std::vector<double> grad_constraints(constraints.size() * numVariables);
    for(unsigned int i = 0; i < grad_objective.size(); i++){
      grad_objective[i] = 2.0*sx[i]/**fscale*/;
      grad_constraints[i] = -2.0*px[i];
      grad_constraints[numVariables+i] = -2.0*qx[i];
    }

    std::cout << "iteration: " << it << "  "
              << "objective: " << objective << "  "
              << "constraints[0]: " << constraints[0] << "  "
              << "constraints[1]: " << constraints[1] << std::endl;

    // call MMA and update the current point
    conveks::mma::updateCurrentPoint(objective, grad_objective, constraints, grad_constraints);
    conveks::mma::getOuterIteration(it);
  }

  // This should be called at the end of MMA 
  conveks::mma::finalize();

  // This should be called at the end
  conveks::finalize();
}